"""
This file is auto generated. Do not edit this file manually.
Source code: [monorepo]/contracts/generator.py
Last generated: 2024-02-12 18:10:07.427816
"""

import typing
from dataclasses import dataclass

CommonUserId = str

CommonUserEmail = str

CommonUserRolesItem = typing.Literal["workflow_viewer"]

CommonUserRolesItemValues: typing.List[CommonUserRolesItem] = ["workflow_viewer"]

CommonUserRoles = typing.List[CommonUserRolesItem]

CommonUserProjectId = str

CommonUserCreatedAt = str


@dataclass
class CommonUser:
    id: CommonUserId
    email: CommonUserEmail
    roles: CommonUserRoles
    project_id: CommonUserProjectId
    created_at: CommonUserCreatedAt

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "email": self.email,
            "roles": self.roles,
            "projectId": self.project_id,
            "createdAt": self.created_at,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "CommonUser":
        return CommonUser(
            id=str(data["id"]),
            email=str(data["email"]),
            roles=[item for item in data["roles"]],
            project_id=str(data["projectId"]),
            created_at=str(data["createdAt"]),
        )


CloudApiCliUserGetResponse = CommonUser

CloudApiCliAuthInfoResponseAuthorId = str

CloudApiCliAuthInfoResponseProjectId = str


@dataclass
class CloudApiCliAuthInfoResponse:
    author_id: CloudApiCliAuthInfoResponseAuthorId
    project_id: CloudApiCliAuthInfoResponseProjectId

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "authorId": self.author_id,
            "projectId": self.project_id,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "CloudApiCliAuthInfoResponse":
        return CloudApiCliAuthInfoResponse(
            author_id=str(data["authorId"]),
            project_id=str(data["projectId"]),
        )


CloudApiCliBuildCreateResponseUrl = str

CloudApiCliBuildCreateResponseBuildId = str


@dataclass
class CloudApiCliBuildCreateResponse:
    url: CloudApiCliBuildCreateResponseUrl
    build_id: CloudApiCliBuildCreateResponseBuildId

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "url": self.url,
            "buildId": self.build_id,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any]
    ) -> "CloudApiCliBuildCreateResponse":
        return CloudApiCliBuildCreateResponse(
            url=str(data["url"]),
            build_id=str(data["buildId"]),
        )


AbstraLibApiStageCardContentKey = str

AbstraLibApiStageCardContentValue = typing.Dict[str, typing.Any]

AbstraLibApiStageCardContentType = str


@dataclass
class AbstraLibApiStageCardContent:
    key: AbstraLibApiStageCardContentKey
    value: AbstraLibApiStageCardContentValue
    type: AbstraLibApiStageCardContentType

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "key": self.key,
            "value": self.value,
            "type": self.type,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStageCardContent":
        return AbstraLibApiStageCardContent(
            key=str(data["key"]),
            value=dict(**data["value"]),
            type=str(data["type"]),
        )


AbstraLibApiStageCardContentsItem = AbstraLibApiStageCardContent

AbstraLibApiStageCardContents = typing.List[AbstraLibApiStageCardContentsItem]

AbstraLibApiVisualizationName = str

AbstraLibApiVisualizationType = typing.Union[
    typing.Literal["text"],
    typing.Literal["image"],
    typing.Literal["table"],
    typing.Literal["progress-bar"],
]

AbstraLibApiVisualizationTypeValues: typing.List[AbstraLibApiVisualizationType] = [
    "text",
    "image",
    "table",
    "progress-bar",
]


@dataclass
class AbstraLibApiVisualization:
    name: AbstraLibApiVisualizationName
    type: AbstraLibApiVisualizationType

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "name": self.name,
            "type": self.type,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiVisualization":
        return AbstraLibApiVisualization(
            name=str(data["name"]),
            type=data["type"],
        )


AbstraLibApiStageRunId = str

AbstraLibApiStageRunStage = str

AbstraLibApiStageRunData = typing.Dict[str, typing.Any]

AbstraLibApiStageRunStatus = typing.Union[
    typing.Literal["waiting"],
    typing.Literal["running"],
    typing.Literal["finished"],
    typing.Literal["failed"],
    typing.Literal["abandoned"],
]

AbstraLibApiStageRunStatusValues: typing.List[AbstraLibApiStageRunStatus] = [
    "waiting",
    "running",
    "finished",
    "failed",
    "abandoned",
]

AbstraLibApiStageRunCreatedAt = str

AbstraLibApiStageRunParentId = str

AbstraLibApiStageRunExecutionId = str


@dataclass
class AbstraLibApiStageRun:
    id: AbstraLibApiStageRunId
    stage: AbstraLibApiStageRunStage
    data: AbstraLibApiStageRunData
    status: AbstraLibApiStageRunStatus
    created_at: AbstraLibApiStageRunCreatedAt
    parent_id: typing.Optional[AbstraLibApiStageRunParentId]
    execution_id: typing.Optional[AbstraLibApiStageRunExecutionId]

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "stage": self.stage,
            "data": self.data,
            "status": self.status,
            "created_at": self.created_at,
            "parent_id": self.parent_id,
            "execution_id": self.execution_id,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStageRun":
        return AbstraLibApiStageRun(
            id=str(data["id"]),
            stage=str(data["stage"]),
            data=dict(**data["data"]),
            status=data["status"],
            created_at=str(data["created_at"]),
            parent_id=str(data.get("parent_id")),
            execution_id=str(data.get("execution_id")),
        )


AbstraLibApiStageRunCardId = str

AbstraLibApiStageRunCardCreatedAt = str

AbstraLibApiStageRunCardAssignee = str

AbstraLibApiStageRunCardStatus = AbstraLibApiStageRunStatus

AbstraLibApiStageRunCardContent = AbstraLibApiStageCardContents


@dataclass
class AbstraLibApiStageRunCard:
    id: AbstraLibApiStageRunCardId
    created_at: AbstraLibApiStageRunCardCreatedAt
    assignee: typing.Optional[AbstraLibApiStageRunCardAssignee]
    status: AbstraLibApiStageRunCardStatus
    content: AbstraLibApiStageRunCardContent

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "created_at": self.created_at,
            "assignee": self.assignee,
            "status": self.status,
            "content": self.content,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStageRunCard":
        return AbstraLibApiStageRunCard(
            id=str(data["id"]),
            created_at=str(data["created_at"]),
            assignee=str(data.get("assignee")),
            status=data["status"],
            content=[
                AbstraLibApiStageCardContent.from_dict(item) for item in data["content"]
            ],
        )


AbstraLibApiStageRunCardsItem = AbstraLibApiStageRunCard

AbstraLibApiStageRunCards = typing.List[AbstraLibApiStageRunCardsItem]

AbstraLibApiStageId = str

AbstraLibApiStageType = typing.Union[
    typing.Literal["form"],
    typing.Literal["hook"],
    typing.Literal["job"],
    typing.Literal["script"],
    typing.Literal["iterator"],
    typing.Literal["condition"],
]

AbstraLibApiStageTypeValues: typing.List[AbstraLibApiStageType] = [
    "form",
    "hook",
    "job",
    "script",
    "iterator",
    "condition",
]

AbstraLibApiStageTitle = str

AbstraLibApiStagePath = str

AbstraLibApiStageCanBeStarted = bool


@dataclass
class AbstraLibApiStage:
    id: AbstraLibApiStageId
    type: AbstraLibApiStageType
    title: AbstraLibApiStageTitle
    path: typing.Optional[AbstraLibApiStagePath]
    can_be_started: AbstraLibApiStageCanBeStarted

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "id": self.id,
            "type": self.type,
            "title": self.title,
            "path": self.path,
            "can_be_started": self.can_be_started,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiStage":
        return AbstraLibApiStage(
            id=str(data["id"]),
            type=data["type"],
            title=str(data["title"]),
            path=str(data.get("path")),
            can_be_started=bool(data["can_be_started"]),
        )


AbstraLibApiKanbanColumnSelectedStage = AbstraLibApiStage

AbstraLibApiKanbanColumnStageRunCards = AbstraLibApiStageRunCards

AbstraLibApiStagesItem = AbstraLibApiStage

AbstraLibApiStages = typing.List[AbstraLibApiStagesItem]

AbstraLibApiKanbanColumnStages = AbstraLibApiStages

AbstraLibApiKanbanColumnTotalCount = int


@dataclass
class AbstraLibApiKanbanColumn:
    selected_stage: AbstraLibApiKanbanColumnSelectedStage
    stage_run_cards: AbstraLibApiKanbanColumnStageRunCards
    stages: AbstraLibApiKanbanColumnStages
    total_count: AbstraLibApiKanbanColumnTotalCount

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "selected_stage": self.selected_stage,
            "stage_run_cards": self.stage_run_cards,
            "stages": self.stages,
            "total_count": self.total_count,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiKanbanColumn":
        return AbstraLibApiKanbanColumn(
            selected_stage=AbstraLibApiStage.from_dict(data["selected_stage"]),
            stage_run_cards=[
                AbstraLibApiStageRunCard.from_dict(item)
                for item in data["stage_run_cards"]
            ],
            stages=[AbstraLibApiStage.from_dict(item) for item in data["stages"]],
            total_count=int(data["total_count"]),
        )


AbstraLibApiKanbanColumnsItem = AbstraLibApiKanbanColumn

AbstraLibApiKanbanColumns = typing.List[AbstraLibApiKanbanColumnsItem]

AbstraLibApiKanbanDataColumns = AbstraLibApiKanbanColumns

AbstraLibApiKanbanDataNextStageOptions = AbstraLibApiStages


@dataclass
class AbstraLibApiKanbanData:
    columns: AbstraLibApiKanbanDataColumns
    next_stage_options: AbstraLibApiKanbanDataNextStageOptions

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "columns": self.columns,
            "next_stage_options": self.next_stage_options,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiKanbanData":
        return AbstraLibApiKanbanData(
            columns=[
                AbstraLibApiKanbanColumn.from_dict(item) for item in data["columns"]
            ],
            next_stage_options=[
                AbstraLibApiStage.from_dict(item) for item in data["next_stage_options"]
            ],
        )


AbstraLibApiVisualizationsItem = AbstraLibApiVisualization

AbstraLibApiVisualizations = typing.List[AbstraLibApiVisualizationsItem]

AbstraLibApiEditorLintersRuleName = str

AbstraLibApiEditorLintersRuleLabel = str

AbstraLibApiEditorLintersRuleType = typing.Literal["bug"]

AbstraLibApiEditorLintersRuleTypeValues: typing.List[
    AbstraLibApiEditorLintersRuleType
] = ["bug"]

AbstraLibApiEditorLintersIssueLabel = str

AbstraLibApiEditorLintersFixName = str

AbstraLibApiEditorLintersFixLabel = str


@dataclass
class AbstraLibApiEditorLintersFix:
    name: AbstraLibApiEditorLintersFixName
    label: AbstraLibApiEditorLintersFixLabel

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "name": self.name,
            "label": self.label,
        }

    @staticmethod
    def from_dict(data: typing.Dict[str, typing.Any]) -> "AbstraLibApiEditorLintersFix":
        return AbstraLibApiEditorLintersFix(
            name=str(data["name"]),
            label=str(data["label"]),
        )


AbstraLibApiEditorLintersIssueFixesItem = AbstraLibApiEditorLintersFix

AbstraLibApiEditorLintersIssueFixes = typing.List[
    AbstraLibApiEditorLintersIssueFixesItem
]


@dataclass
class AbstraLibApiEditorLintersIssue:
    label: AbstraLibApiEditorLintersIssueLabel
    fixes: AbstraLibApiEditorLintersIssueFixes

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "label": self.label,
            "fixes": self.fixes,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any]
    ) -> "AbstraLibApiEditorLintersIssue":
        return AbstraLibApiEditorLintersIssue(
            label=str(data["label"]),
            fixes=[
                AbstraLibApiEditorLintersFix.from_dict(item) for item in data["fixes"]
            ],
        )


AbstraLibApiEditorLintersRuleIssuesItem = AbstraLibApiEditorLintersIssue

AbstraLibApiEditorLintersRuleIssues = typing.List[
    AbstraLibApiEditorLintersRuleIssuesItem
]


@dataclass
class AbstraLibApiEditorLintersRule:
    name: AbstraLibApiEditorLintersRuleName
    label: AbstraLibApiEditorLintersRuleLabel
    type: AbstraLibApiEditorLintersRuleType
    issues: AbstraLibApiEditorLintersRuleIssues

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "name": self.name,
            "label": self.label,
            "type": self.type,
            "issues": self.issues,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any]
    ) -> "AbstraLibApiEditorLintersRule":
        return AbstraLibApiEditorLintersRule(
            name=str(data["name"]),
            label=str(data["label"]),
            type=data["type"],
            issues=[
                AbstraLibApiEditorLintersIssue.from_dict(item)
                for item in data["issues"]
            ],
        )


AbstraLibApiEditorLintersFixResponseSuccess = bool


@dataclass
class AbstraLibApiEditorLintersFixResponse:
    success: AbstraLibApiEditorLintersFixResponseSuccess

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "success": self.success,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any]
    ) -> "AbstraLibApiEditorLintersFixResponse":
        return AbstraLibApiEditorLintersFixResponse(
            success=bool(data["success"]),
        )


AbstraLibApiEditorLintersCheckResponseItem = AbstraLibApiEditorLintersRule

AbstraLibApiEditorLintersCheckResponse = typing.List[
    AbstraLibApiEditorLintersCheckResponseItem
]

AbstraLibApiEditorEnvVarsModelKey = str

AbstraLibApiEditorEnvVarsModelValue = str


@dataclass
class AbstraLibApiEditorEnvVarsModel:
    key: AbstraLibApiEditorEnvVarsModelKey
    value: AbstraLibApiEditorEnvVarsModelValue

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        return {
            "key": self.key,
            "value": self.value,
        }

    @staticmethod
    def from_dict(
        data: typing.Dict[str, typing.Any]
    ) -> "AbstraLibApiEditorEnvVarsModel":
        return AbstraLibApiEditorEnvVarsModel(
            key=str(data["key"]),
            value=str(data["value"]),
        )


AbstraLibApiEditorEnvVarsListResponseItem = AbstraLibApiEditorEnvVarsModel

AbstraLibApiEditorEnvVarsListResponse = typing.List[
    AbstraLibApiEditorEnvVarsListResponseItem
]

AbstraLibApiEditorEnvVarsCreateRequest = AbstraLibApiEditorEnvVarsModel

AbstraLibApiEditorEnvVarsCreateResponse = AbstraLibApiEditorEnvVarsModel



from werkzeug.datastructures import FileStorage
from typing import Any, Dict, List, Optional, Tuple
import pkg_resources
import flask, pkgutil, webbrowser
from pathlib import Path

from abstra_internals.server.controller.linters import check_linters
from ...utils.validate import validate_json
from ...cloud_api import get_ai_messages, get_auth_info, get_project_info
from ...utils import path2module, module2path, files_from_directory
from ...widgets.apis import get_random_filepath, internal_path
from ...execution.execution import Execution
from ...interface.cli.deploy import deploy
from ...settings import Settings
from ...repositories import (
    requirements_repository,
    stage_run_repository,
    execution_logs_repository,
    execution_repository,
)
from ...repositories.execution import ExecutionRepository
from ...repositories.execution_logs import ExecutionLogsRepository
from ...repositories.stage_run import StageRunRepository

from ...debugger.vscode import (
    is_launch_json_configured,
    is_settings_json_set,
    configure_settings_json,
    configure_launch_json,
    is_client_connected,
    start_debugger,
)

from ...credentials import (
    delete_credentials,
    get_credentials,
    resolve_headers,
    set_credentials,
)

from ...templates import (
    ensure_abstraignore,
    new_script_code,
    new_form_code,
    new_hook_code,
    new_job_code,
)

from ...repositories.project.project import (
    ProjectRepository,
    StyleSettings,
    ActionStage,
    ScriptStage,
    ScriptStage,
    ScriptStage,
    FormStage,
    HookStage,
    JobStage,
)


class UnknownNodeTypeError(Exception):
    def __init__(self, node_type: str):
        self.node_type = node_type

    def __str__(self):
        return f"Unknown node type {self.node_type}"


class SelfTransitionError(Exception):
    def __init__(self, node_type: str, node_id: str):
        self.node_type = node_type
        self.node_id = node_id

    def __str__(self):
        return f"You can't add a transition to itself."


class TransitionToJobError(Exception):
    def __init__(
        self, source_type: str, source_id: str, target_type: str, target_id: str
    ):
        self.source_type = source_type
        self.source_id = source_id
        self.target_type = target_type
        self.target_id = target_id

    def __str__(self):
        return f"You can't add a transition to a job. Use a script instead."


class DoubleTransitionError(Exception):
    def __init__(
        self, source_type: str, source_id: str, target_type: str, target_id: str
    ):
        self.source_type = source_type
        self.source_id = source_id
        self.target_type = target_type
        self.target_id = target_id

    def __str__(self):
        return f"You can't add the same transition twice."


class MainController:
    stage_run_repository: StageRunRepository
    execution_repository: ExecutionRepository
    execution_logs_repository: ExecutionLogsRepository

    def __init__(self):
        if not ProjectRepository.exists():
            ProjectRepository.initialize()
        requirements = requirements_repository.load()
        requirements.ensure("abstra", pkg_resources.get_distribution("abstra").version)
        requirements_repository.save(requirements)
        ensure_abstraignore(Settings.root_path)

        self.stage_run_repository = stage_run_repository
        self.execution_repository = execution_repository
        self.execution_logs_repository = execution_logs_repository

    def deploy(self):
        rules = check_linters()

        issues = []
        for rule in rules:
            issues += rule.issues

        if len(issues) > 0:
            raise Exception(
                "Please fix all linter issues before deploying your project."
            )

        deploy()

    def get_workspace(self) -> StyleSettings:
        project = ProjectRepository.load()
        return project.workspace

    def init_code_file(self, path: str, code: str):
        Settings.root_path.joinpath(path).write_text(code, encoding="utf-8")

    def open_file(self, file_path: str, mode: str, create_if_not_exists: bool = False):
        if mode == "module" or mode == "package":
            file_path = str(module2path(file_path, mode == "package"))
        complete_file_path = Settings.root_path.joinpath(file_path)

        if create_if_not_exists and not complete_file_path.is_file():
            complete_file_path.touch()

        webbrowser.open(complete_file_path.absolute().as_uri())

    def read_file(self, file: str):
        file_path = Settings.root_path.joinpath(file)
        if not file_path.is_file():
            return None
        return Settings.root_path.joinpath(file).read_text(encoding="utf-8")

    def check_file(self, file_path: str):
        return Settings.root_path.joinpath(file_path).is_file()

    def list_files(self, path: str = ".", mode: str = "file"):
        parent_path = Settings.root_path.joinpath(path)
        if mode in ["file", "image", "python-file"]:
            if mode == "image":
                allowed_suffixes = [
                    ".png",
                    ".jpg",
                    ".jpeg",
                    ".gif",
                    ".svg",
                    ".webp",
                    ".jfif",
                    ".pjp",
                    ".pjpeg",
                ]
            elif mode == "python-file":
                allowed_suffixes = [".py"]
            else:
                allowed_suffixes = None

            return [
                dict(
                    name=str(file.relative_to(parent_path)),
                    path=str(file.relative_to(Settings.root_path)),
                    type="file" if file.is_file() else "dir",
                )
                for file in files_from_directory(parent_path)
                if file.is_dir()
                or not allowed_suffixes
                or file.suffix in allowed_suffixes
            ]

        elif mode == "module":
            return [
                dict(
                    name=name,
                    path=path2module(Path(path).joinpath(name)),
                    type="package" if ispkg else "module",
                )
                for module_finder, name, ispkg in pkgutil.iter_modules(
                    [str(parent_path)]
                )
            ]

    def update_workspace(self, changes: Dict[str, Any]):
        project = ProjectRepository.load()
        project.workspace.update(changes, project.forms)
        ProjectRepository.save(project)
        return project.workspace

    def is_initial(self, id: str):
        project = ProjectRepository.load()
        stage = project.get_action(id)
        if not stage:
            raise Exception(f"Stage {id} not found")
        return project.is_initial(stage)

    def create_script(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> ScriptStage:
        project = ProjectRepository.load()
        script = ScriptStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(script.file, new_script_code)
        project.add_stage(script)
        ProjectRepository.save(project)

        return script

    def get_scripts(self) -> List[ScriptStage]:
        project = ProjectRepository.load()
        return project.scripts

    def get_script(self, id: str) -> Optional[ScriptStage]:
        project = ProjectRepository.load()
        return project.get_script(id)

    def delete_script(self, id: str, remove_file: bool = False):
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    def create_form(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> FormStage:
        project = ProjectRepository.load()
        form = FormStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(form.file, new_form_code)
        project.add_stage(form)
        ProjectRepository.save(project)
        return form

    def get_forms(self) -> List[FormStage]:
        project = ProjectRepository.load()
        return project.forms

    def get_form(self, id: str) -> Optional[FormStage]:
        project = ProjectRepository.load()
        return project.get_form(id)

    def get_form_by_path(self, path: str) -> Optional[FormStage]:
        project = ProjectRepository.load()
        return project.get_form_by_path(path)

    def write_test_data(self, data: str) -> None:
        if not validate_json(data):
            raise Exception("Invalid JSON")
        test_file = Settings.root_path / ".abstra" / "test_data.json"
        test_file.write_text(data, encoding="utf-8")

    def read_test_data(self) -> str:
        test_file = Settings.root_path / ".abstra" / "test_data.json"
        if not test_file.is_file():
            return "{}"
        return test_file.read_text(encoding="utf-8")

    def delete_form(self, id: str, remove_file: bool = False):
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    def create_hook(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> HookStage:
        project = ProjectRepository.load()
        hook = HookStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(hook.file, new_hook_code)
        project.add_stage(hook)
        ProjectRepository.save(project)
        return hook

    def get_hook(self, id: str) -> Optional[HookStage]:
        project = ProjectRepository.load()
        return project.get_hook(id)

    def get_hooks(self) -> List[HookStage]:
        project = ProjectRepository.load()
        return project.hooks

    def get_hook_by_path(self, path: str) -> Optional[HookStage]:
        project = ProjectRepository.load()
        return project.get_hook_by_path(path)

    def delete_hook(self, id: str, remove_file: bool = False) -> None:
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    def get_jobs(self) -> List[JobStage]:
        project = ProjectRepository.load()
        return project.jobs

    def get_job(self, id: str) -> Optional[JobStage]:
        project = ProjectRepository.load()
        stage = project.get_stage(id)

        if isinstance(stage, JobStage):
            return stage

        return None

    def create_job(
        self, title: str, file: str, workflow_position: Tuple[int, int] = (0, 0)
    ) -> JobStage:
        project = ProjectRepository.load()
        job = JobStage.create(title, file, workflow_position=workflow_position)
        self.init_code_file(job.file, new_job_code)
        project.add_stage(job)
        ProjectRepository.save(project)
        return job

    def update_stage(self, id: str, changes: Dict[str, Any]) -> ActionStage:
        project = ProjectRepository.load()
        stage = project.get_action(id)

        if not stage:
            raise Exception(f"Stage with id {id} not found")

        code_content = changes.get("code_content")
        if code_content:
            Settings.root_path.joinpath(stage.file_path).write_text(
                code_content, encoding="utf-8"
            )
            del changes["code_content"]

        test_data = changes.get("test_data")
        if test_data:
            self.write_test_data(test_data)
            del changes["test_data"]

        stage = project.update_stage(stage, changes)
        ProjectRepository.save(project)
        return stage

    def delete_job(self, id: str, remove_file: bool = False):
        project = ProjectRepository.load()
        project.delete_stage(id, remove_file)
        ProjectRepository.save(project)

    # Login
    def get_credentials(self):
        return get_credentials()

    def get_login(self):
        headers = resolve_headers()
        if not headers:
            return {"logged": False, "reason": "NO_API_TOKEN"}
        return get_auth_info(headers)

    def create_login(self, token):
        set_credentials(token)
        return self.get_login()

    def delete_login(self):
        delete_credentials()
        return self.get_login()

    # Project
    def get_project_info(self):
        headers = resolve_headers()
        if headers is None:
            flask.abort(401)
        return get_project_info(headers)

    # AI
    def send_ai_message(self, messages, stage):
        headers = resolve_headers() or {}
        yield from get_ai_messages(messages, stage, headers)

    # files
    def save_file(self, file: FileStorage, filename: str):
        name, path = get_random_filepath(filename)
        file.save(path)
        return name

    def get_file(self, path: str):
        return internal_path(path)

    def abort_execution(self, id: str):
        execution = Execution.get_execution_by_id(id)
        if execution:
            # execution.stop()
            pass

    # Debugger
    def get_debugger_status(self):
        return dict(
            port=start_debugger(),
            connected=is_client_connected(),
            is_launch_json_configured=is_launch_json_configured()
            and is_settings_json_set(),
        )

    def create_vscode_launch(self):
        configure_launch_json()
        configure_settings_json()



from unittest import TestCase
from tests.fixtures import init_dir, clear_dir
from .env_vars import env_vars_repository_factory
from ..contracts_generated import AbstraLibApiEditorEnvVarsModel
from ..repositories.project.project import ProjectRepository, ScriptStage


class TestEnvVarsRepository(TestCase):
    def setUp(self):
        self.dir = init_dir()
        self.env_var_repo = env_vars_repository_factory()

    def tearDown(self):
        clear_dir(self.dir)

    def test_load_default_empty(self):
        self.assertEqual(self.env_var_repo.list(), [])

    def test_load(self):
        var = AbstraLibApiEditorEnvVarsModel(key="key", value="value")
        self.env_var_repo.set(var)
        self.assertEqual(self.env_var_repo.list(), [var])

    def test_set(self):
        var = AbstraLibApiEditorEnvVarsModel(key="key", value="value")
        self.env_var_repo.set(var)
        self.assertEqual(self.env_var_repo.list(), [var])

    def test_unset(self):
        var = AbstraLibApiEditorEnvVarsModel(key="key", value="value")
        self.env_var_repo.set(var)
        self.env_var_repo.unset(var)
        self.assertEqual(self.env_var_repo.list(), [])

    def test_check_no_file_is_valid(self):
        self.assertTrue(self.env_var_repo.check())

    def test_check_empty_file_is_valid(self):
        open(".env", "w").close()
        self.assertTrue(self.env_var_repo.check())

    def test_check_multiple_lines_is_valid(self):
        with open(".env", "w") as f:
            f.write("key=value\nkey2=value2")
        self.assertTrue(self.env_var_repo.check())

    def test_check_no_value_is_valid(self):
        with open(".env", "w") as f:
            f.write("key=value\nkey2")
        self.assertTrue(self.env_var_repo.check())

    def test_check_invalid_returns_false(self):
        with open(".env", "w") as f:
            f.write("===")
        self.assertFalse(self.env_var_repo.check())

    def test_env_vars_in_code(self):
        project = ProjectRepository.load()
        code_path = self.dir / "code.py"
        code_content = "\n".join(["import os", "os.getenv('FOO')", "os.environ['BAR']"])
        code_path.write_text(code_content)
        script = ScriptStage(
            file=str(code_path),
            id="script",
            is_initial=False,
            title="foo bar",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(script)
        ProjectRepository.save(project)
        env_vars_in_code = self.env_var_repo.get_env_vars_in_code()
        self.assertEqual(env_vars_in_code.keys(), {"FOO", "BAR"})



import sys
from typing import Callable, Dict, List, Optional, Union

from ....contract import forms_contract
from ....widgets.prop_check import validate_widget_props
from .connection import close, receive, send
from .generated.widget_schema import WidgetSchema
from .page_response import PageResponse
from .reactive import Reactive


class Page(WidgetSchema):
    """A form page that can be displayed to the user

    This is a page that can be displayed to the user. It can be used to
    show data as well as collect informations. After configuring the
    inputs and outputs, use the run method to display the form to the
    user and collect the answers.
    """

    def __init__(
        self,
        actions: Union[str, List[str]] = "i18n_next_action",
        validate: Optional[Callable] = None,
        end_program: bool = False,
        reactive_polling_interval=0,
        context: Optional[Dict] = None,
    ):
        super().__init__()
        self.__actions = actions
        self.__validate = validate
        self.__end_program = end_program
        self.__reactive_polling_interval = reactive_polling_interval
        self.__context = context or {}

    def run(
        self,
        actions: Union[str, List[str]] = "i18n_next_action",
        validate: Optional[Callable] = None,
        end_program: bool = False,
        reactive_polling_interval=0,
        context: Optional[Dict] = None,
        steps_info: Optional[Dict] = None,
    ) -> PageResponse:
        """Run the form

        Args:
            button_text: The text of the button that is used to submit the form
            end_program: End program whilst showing this page, making this an end page
            reactive_polling_interval: Interval to poll and rerender reactive slots, in seconds (0 to disable)

        Returns:
            The form result as a dict with the keys being the key of the input and the value being the value of the input
        """

        actions = actions if actions != "i18n_next_action" else self.__actions
        end_program = end_program if end_program != False else self.__end_program
        reactive_polling_interval = (
            reactive_polling_interval
            if reactive_polling_interval != 0
            else self.__reactive_polling_interval
        )
        self.__context = context if context is not None else self.__context
        validate = validate if validate != None else self.__validate

        if self.__context:
            for widget in self.widgets:
                if (
                    not isinstance(widget, Reactive)
                    and hasattr(widget, "key")
                    and widget.key in self.__context
                ):
                    widget.value = self.__context[widget.key]

        for widget in self.widgets:
            if hasattr(widget, "key"):
                self.__context[widget.key] = widget.value

        rendered_page = self.render(context=self.__context)

        self.__check_widget_props(rendered_page)

        if self.__is_progress_screen():
            self.__send_mount_page_message(
                widgets=rendered_page,
                actions=[],
                end_program=end_program,
                reactive_polling_interval=reactive_polling_interval,
                steps_info=steps_info,
            )
            return PageResponse({}, "")

        self.__send_mount_page_message(
            widgets=rendered_page,
            actions=self.__actions_property(actions, end_program),
            end_program=end_program,
            reactive_polling_interval=reactive_polling_interval,
            steps_info=steps_info,
        )
        response: Dict = self.__handle_page_user_events(validate_func=validate)

        if end_program:
            close()
            sys.exit(0)

        return PageResponse(
            self.parse_value(response["payload"]),
            response.get("action"),
        )

    def __handle_page_user_events(self, validate_func: Optional[Callable]):
        while True:
            response = receive()

            self.update(response["payload"])

            # TODO: handle back actions without string comparison
            if response.get("action") == "i18n_back_action":
                break

            # TODO: Refactor validation to use values instead of payload
            parsed_payload = self.parse_value(response["payload"])
            validation_result = self._get_validation_result(
                validation=validate_func, payload=parsed_payload
            )

            if response["type"] != "form:user-event":
                self.set_errors()
                if not self.has_errors() and validation_result.get("status"):
                    break

            rendered_page = self.render(self.__context)

            self.__send_form_update_message(
                widgets=rendered_page,
                validation=validation_result,
                event_seq=response["seq"],
            )

        return response

    def __check_widget_props(self, rendered_page):
        for widget in rendered_page:
            if isinstance(widget, list):
                self.__check_widget_props(widget)
            else:
                validate_widget_props(widget)

    def __actions_property(self, actions, end_page):
        if end_page:
            return []
        elif isinstance(actions, list):
            return actions
        elif actions is None:
            return []
        return [actions]

    def __is_progress_screen(self):
        return len(self.widgets) == 1 and self.widgets[0].type == "progress-output"

    def update(self, payload):
        parsed_values = self.parse_value(payload)
        self.set_values(parsed_values)
        self.__context.update(parsed_values)

    def _get_validation_result(
        self, validation, payload
    ) -> forms_contract.ValidationResult:
        # TODO: Refactor this to use widget.get_value() instead of payload
        validation_status = True
        validation_message = ""

        if validation:
            validation_response = validation(payload)
            if type(validation_response) == bool:
                validation_status = validation_response
                validation_message = ""
            elif type(validation_response) == str:
                validation_status = False
                validation_message = validation_response

        return {"status": validation_status, "message": validation_message}

    def __send_mount_page_message(
        self,
        widgets: list,
        actions: list,
        end_program: bool,
        reactive_polling_interval,
        steps_info,
    ):
        send(
            forms_contract.FormMountPageMessage(
                widgets=widgets,
                actions=actions,
                end_program=end_program,
                reactive_polling_interval=reactive_polling_interval,
                steps=steps_info,
            )
        )

    def __send_form_update_message(
        self, widgets: list, validation: forms_contract.ValidationResult, event_seq: int
    ):
        send(
            forms_contract.FormUpdatePageMessage(
                widgets=widgets,
                validation=validation,
                event_seq=event_seq,
            )
        )



from ..widget_base import Output
from typing import Any
import json


class PandasOutput(Output):
    type = "pandas-output"

    def __init__(self, df: Any = None, **kwargs):
        self.set_props(dict(df=df, **kwargs))

    def set_props(self, props):
        self.df = props.get("df", "")
        self.full_width = props.get("full_width", False)
        self.display_index = props.get("display_index", False)
        self.label = props.get("label", "")
        self.actions = props.get("actions", [])
        self.filterable = props.get("filterable", False)
        self.editable = props.get("editable", False)

    def serialize_table(self):
        if self.df is None:
            import pandas as pd

            df = pd.DataFrame({"change the": [1, 2, 3], "df property": [4, 5, 6]})
            df_json = df.to_json(orient="table")
            if not isinstance(df_json, str):
                raise Exception("df.to_json() did not return a string")
            serialized = json.loads(df_json)
            del serialized["schema"]["pandas_version"]
            return serialized
        serialized = json.loads(self.df.to_json(orient="table"))
        del serialized["schema"]["pandas_version"]
        return serialized

    def render(self, ctx: dict):
        return {
            "type": self.type,
            "table": self.serialize_table(),
            "fullWidth": self.full_width,
            "displayIndex": self.display_index,
            "label": self.label,
            "actions": self.actions,
            "filterable": self.filterable,
            "editable": self.editable,
        }



import abstra.forms as af

"""
Abstra forms are the simplest way to build user interfaces for your workflows.
"""

name = af.read("Hello there! What is your name?")
af.display(f"Hello {name}!")


ans = af.read_multiple_choice("Are you familiar with Abstra?", ["Yes", "No"])

if ans == "Yes":
    af.display("Great! Have fun!")
else:
    af.display_link(
        "https://docs.abstra.io/forms/overview", link_text="Check out the docs!"
    )



from unittest import TestCase
from .unset_get_data import UnsetGetData
from tests.fixtures import init_dir, clear_dir
from ...repositories.project.project import ProjectRepository, ScriptStage


class UnsetGetDataTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_unset_get_data_valid_default(self):
        rule = UnsetGetData()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_unset_get_data_invalid(self):
        rule = UnsetGetData()
        project = ProjectRepository.load()
        code_path = self.root / "test.py"
        code_path.write_text(
            "\n".join(
                ["from abstra_internals.workflows import get_data", "get_data('test')"]
            )
        )
        script = ScriptStage(
            id="test",
            file=str(code_path),
            title="Test",
            is_initial=False,
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.add_stage(script)
        ProjectRepository.save(project)
        self.assertEqual(len(rule.find_issues()), 1)



import datetime
import threading
import uuid
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Literal, Optional, Tuple, Union

from .stage_run_manager import StageRunManager

from ..compatibility import compat_traceback
from ..modules import import_as_new
from ..repositories.project.project import ActionStage
from ..repositories.stage_run import (
    StageRun,
)

from ..repositories.execution import (
    ExecutionRepository,
    ExecutionStatus,
    ExecutionDTO,
)

from ..repositories.execution_logs import (
    ExecutionLogsRepository,
    StdioLogEntry,
    UnhandledExceptionLogEntry,
)
from ..repositories.project.project import ActionStage
from ..repositories.stage_run import StageRun
from ..utils.datetime import to_utc_iso_string


class NoExecutionFound(Exception):
    def __str__(self) -> str:
        return "No execution found. If you are running this locally, make sure you are running it with `abstra serve`. For more information, see https://docs.abstra.io"


ABSTRA_RUN_KEY = "abstra-run-id"


@dataclass
class RequestData:
    body: str
    method: str
    query_params: Dict[str, str]
    headers: Dict[str, str]

    def sanitize_dict(self, headers: Dict[str, Any]):
        sanitized: Dict[str, str] = {}

        for key, value in headers.items():
            if isinstance(value, str):
                sanitized[key] = value

        return sanitized

    def to_dto(self):
        return dict(
            headers=self.sanitize_dict(self.headers),
            body=self.body,
            method=self.method,
            queryParams=self.query_params,
        )


ResponseData = Tuple[Union[str, Dict], int, Dict]


@dataclass
class ExecutionContext:
    request: RequestData
    response: Optional[ResponseData]

    def to_dto(self):
        return dict(
            request=self.request.to_dto(),
            response=self.response,
        )


class Execution:
    executions: ClassVar[Dict[int, "Execution"]] = {}

    id: str
    created_at: datetime.datetime
    is_initial: bool
    status: Optional[ExecutionStatus] = None
    stage: ActionStage
    context: ExecutionContext

    stage_run_manager: StageRunManager
    execution_repository: ExecutionRepository
    execution_logs_repository: ExecutionLogsRepository

    _log_count: int = 0

    stage_run_draft: Optional[StageRun] = None
    stage_run_freezed: Optional[StageRun] = None

    @classmethod
    def get_execution_by_id(cls, execution_id: str) -> Optional["Execution"]:
        for execution in cls.executions.values():
            if execution.id == execution_id:
                return execution
        return None

    @classmethod
    def get_current_execution(cls) -> Optional["Execution"]:
        thread_id = threading.get_ident()
        return cls.executions.get(thread_id)

    @property
    def log_count(self) -> int:
        self._log_count += 1
        return self._log_count

    def __init__(
        self,
        stage: ActionStage,
        is_initial: bool,
        request: RequestData,
        stage_run_manager: StageRunManager,
        execution_repository: ExecutionRepository,
        execution_logs_repository: ExecutionLogsRepository,
        execution_id: Optional[str] = None,
    ):
        self.id = execution_id or uuid.uuid4().__str__()
        self.thread_id = threading.get_ident()

        self.stage = stage
        self.is_initial = is_initial
        self.created_at = datetime.datetime.now()

        self.context = ExecutionContext(
            request=request,
            response=None,
        )
        Execution.executions[self.thread_id] = self

        self.stage_run_manager = stage_run_manager
        self.execution_repository = execution_repository
        self.execution_logs_repository = execution_logs_repository

    @property
    def stage_run(self) -> Optional[StageRun]:
        return self.stage_run_draft

    @stage_run.setter
    def stage_run(self, stage_run: StageRun) -> None:
        self.stage_run_freezed = stage_run
        self.stage_run_draft = stage_run.clone()

    def to_dto(self) -> ExecutionDTO:
        if not self.status:
            raise Exception("Internal error: status is not set")

        return ExecutionDTO(
            id=self.id,
            status=self.status,
            created_at=to_utc_iso_string(self.created_at),
            context=self.context.to_dto(),
            stage_id=self.stage.id,
            stage_run_id=self.stage_run.id if self.stage_run else None,
        )

    def save(self) -> None:
        self.execution_repository.update(self.to_dto())

    def stdio(self, event: Literal["stderr", "stdout"], text: str) -> None:
        self.execution_logs_repository.save(
            StdioLogEntry(
                execution_id=self.id,
                event=event,
                payload={"text": text},
                created_at=datetime.datetime.now(),
                sequence=self.log_count,
            )
        )

    def log_exception(self, e: Exception) -> None:
        self.execution_logs_repository.save(
            UnhandledExceptionLogEntry(
                execution_id=self.id,
                payload={
                    "text": "".join(
                        compat_traceback.format_exception(e),
                    ),
                },
                sequence=self.log_count,
                created_at=datetime.datetime.now(),
            )
        )

    @abstractmethod
    def received_stage_run_id(self) -> Optional[str]:
        raise NotImplementedError()

    @abstractmethod
    def handle_start(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def handle_success(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def handle_failure(self, e: Exception) -> None:
        raise NotImplementedError()

    @abstractmethod
    def handle_lock_failed(self) -> None:
        raise NotImplementedError()

    def attempt_handle_exception(self, e: Exception) -> bool:
        return False

    def run(self) -> None:
        self.status = "running"
        self.execution_repository.create(self.to_dto())
        self.handle_start()
        stage_run_id = self.received_stage_run_id()
        self.stage_run = self.stage_run_manager.init_stage_run(self.stage, stage_run_id)

        lock_held = self.set_stage_run_status()

        if self.stage_run and not lock_held:
            self.status = "lock-failed"
            self.handle_lock_failed()
            self.save()
            return

        try:
            try:
                import_as_new(self.stage.file)
            except SystemExit as e:
                had_errors = e.code is not None and e.code != 0
                if had_errors:
                    raise e

            self.status = "finished"
            self.handle_success()
        except Exception as e:
            if not self.attempt_handle_exception(e):
                self.status = "failed"
                self.log_exception(e)
                self.handle_failure(e)
        finally:
            self.save()
            self.unbind()
            self.set_stage_run_status()

    def set_stage_run_status(self) -> bool:
        if not self.stage_run or not self.status:
            return False

        return self.stage_run_manager.change_state(
            self.stage_run.id, self.status, self.id
        )

    def unbind(self) -> None:
        del Execution.executions[self.thread_id]


def get_current_execution_throwable() -> Execution:
    execution = Execution.get_current_execution()
    if not execution:
        raise NoExecutionFound()
    return execution



from abc import ABC, abstractmethod
from typing import List, Optional

from ..repositories.stage_run import (
    StageRun,
    StageRunRepository,
    GetStageRunByQueryFilter,
)

from ..repositories.project.project import ActionStage
from ..repositories.stage_run import StageRun, StageRunRepository


class InvalidStageRunId(Exception):
    pass


class UnsetStageRun(Exception):
    pass


ABSTRA_RUN_KEY = "abstra-run-id"


class StageRunManager(ABC):
    stage_run_repository: StageRunRepository

    def find(self, filter: GetStageRunByQueryFilter) -> List[StageRun]:
        return self.stage_run_repository.find(filter)

    def get(self, stage_run_id: str) -> StageRun:
        return self.stage_run_repository.get(stage_run_id)

    def create_initial(self, stage: str) -> StageRun:
        return self.stage_run_repository.create_initial(stage=stage)

    def change_state(self, stage_run_id: str, status: str, execution_id: str) -> bool:
        return self.stage_run_repository.change_state(
            stage_run_id, status, execution_id
        )

    def get_not_abandoned_stage_run(
        self, stage_run: StageRun, stage_id: str
    ) -> StageRun:
        if stage_run.status != "abandoned":
            return stage_run

        filter = GetStageRunByQueryFilter()
        filter.parent_id = stage_run.id
        filter.stage = stage_id

        next_matches = self.find(filter)

        assert (
            len(next_matches) == 1
        ), "Internal error: abandoned stage run does not have exactly one child"

        return self.get_not_abandoned_stage_run(next_matches[0], stage_id)

    @abstractmethod
    def init_stage_run(
        self, stage: ActionStage, stage_run_id: Optional[str]
    ) -> StageRun:
        raise NotImplementedError()


class AttachedStageRunManager(StageRunManager):
    stage_run_repository: StageRunRepository

    def __init__(
        self,
        stage_run_repository: StageRunRepository,
    ):
        self.stage_run_repository = stage_run_repository

    def init_stage_run(
        self, stage: ActionStage, stage_run_id: Optional[str]
    ) -> StageRun:
        if stage_run_id:
            stage_run_tmp = self.stage_run_repository.get(stage_run_id)
            stage_run = self.get_not_abandoned_stage_run(stage_run_tmp, stage.id)

            if stage_run.stage == stage.id:
                return stage_run

            raise InvalidStageRunId()

        if stage.is_initial:
            return self.stage_run_repository.create_initial(stage=stage.id)

        raise UnsetStageRun()


class DetachedStageRunManager(StageRunManager):
    stage_run_repository: StageRunRepository
    data: dict = {}

    def __init__(
        self,
        stage_run_repository: StageRunRepository,
        data: dict = {},
    ):
        self.stage_run_repository = stage_run_repository
        self.data = data

    def init_stage_run(
        self, stage: ActionStage, stage_run_id: Optional[str]
    ) -> StageRun:
        return self.stage_run_repository.create_initial(stage=stage.id, data=self.data)



import abstra.hooks as ah

"""
Abstra hooks are the simplest way to build endpoints for your workflows.
"""

body, query, headers = ah.get_request()

print("Hook ran!")
print("Received request body:", body)

ah.send_json({"message": "Hello world!"})



import sys
import importlib
import importlib.util as imp_util
from .repositories.project.project import ProjectRepository
from .utils.file import path2module
from pathlib import Path


def import_as_new(filepath: str):
    module_spec = imp_util.spec_from_file_location(filepath, filepath)
    if module_spec is None or module_spec.loader is None:
        raise Exception(f"Could not load module {filepath}")

    module = imp_util.module_from_spec(module_spec)
    if module is None:
        raise Exception(f"Could not load module {filepath}")

    if module_spec.cached is not None:
        Path(module_spec.cached).unlink(missing_ok=True)

    module_spec.loader.exec_module(module)

    return module


def reload_project_local_modules():
    project = ProjectRepository.load()

    for file in project.project_local_dependencies:
        try:
            if not file.exists():
                continue

            module_name = path2module(file)
            module = sys.modules.get(module_name)

            if module is None:
                importlib.import_module(module_name)
            else:
                if module.__spec__ is not None and module.__spec__.cached is not None:
                    Path(module.__spec__.cached).unlink(missing_ok=True)
                importlib.reload(module)
        except Exception as e:
            print(f"Could not reload module from {file} with the following error:")
            print(e)
            continue
